import pandas as pd
import pytest


@pytest.fixture
def dataframe1():
    """Return a simple DataFrame with MultiIndex.

           v0  v1
    i0 i1
    a  c    0   4
       d    1   5
    b  c    2   6
       d    3   7
    """
    return pd.DataFrame(
        {"v0": [0, 1, 2, 3], "v1": [4, 5, 6, 7]},
        index=pd.MultiIndex.from_product([["a", "b"], ["c", "d"]], names=["i0", "i1"]),
    )


@pytest.fixture
def series1():
    """Return a simple Series with MultiIndex.

    i0  i1
    a   c     0
        d     1
    b   c     2
        d     3
    Name: values, dtype: int64
    """
    return pd.Series(
        [0, 1, 2, 3],
        index=pd.MultiIndex.from_product([["a", "b"], ["c", "d"]], names=["i0", "i1"]),
        name="values",
    )


@pytest.fixture
def index1():
    """Return a MultiIndex.

    MultiIndex([('a', 'c'),
                ('a', 'd'),
                ('b', 'c'),
                ('b', 'd')],
               names=['i0', 'i1'])
    """
    return pd.MultiIndex.from_product([["a", "b"], ["c", "d"]], names=["i0", "i1"])


@pytest.fixture
def index2():
    """Return a MultiIndex without names.

    MultiIndex([('a', 'c'),
                ('a', 'd'),
                ('b', 'c'),
                ('b', 'd')],
               )
    """
    return pd.MultiIndex.from_product([["a", "b"], ["c", "d"]])



# Basic test case to test functioning of module's top-level

try:
    import blueetl_core  # noqa

    _top_import_error = None
except Exception as ex:  # pragma: no cover
    _top_import_error = ex


def test_import():
    assert _top_import_error is None



import re

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal, assert_series_equal

from blueetl_core import utils as test_module


@pytest.mark.parametrize(
    "x, expected",
    [
        ([], []),
        ((), []),
        (None, [None]),
        ("", [""]),
        ("a", ["a"]),
        (["a"], ["a"]),
        (("a",), ["a"]),
        ({"a": 1}, [{"a": 1}]),
    ],
)
def test_ensure_list(x, expected):
    result = test_module.ensure_list(x)
    assert result == expected


def test_andor_mask():
    query_list = [
        {"col1": 10},  # select row 1
        {"col1": 11, "col2": 111},  # select row 2
        {"col1": 99, "col2": 100},  # select none
    ]
    df = pd.DataFrame(
        {
            "col1": [0, 10, 11, 20, 21],
            "col2": [100, 110, 111, 111, 111],
        }
    )

    def _filter_func(query):
        return [df[key] == val for key, val in query.items()]

    mask = test_module._and_or_mask(query_list, _filter_func)

    assert_array_equal(mask, [0, 1, 1, 0, 0])


@pytest.mark.parametrize(
    "value, expected_true_indices",
    [
        (10, [1, 6]),
        (-1, []),
        ([11, 20], [2, 4]),
        ({"eq": 10}, [1, 6]),
        ({"ne": 10}, [0, 2, 3, 4, 5, 7]),
        ({"le": 10}, [0, 1, 6, 7]),
        ({"lt": 10}, [0, 7]),
        ({"ge": 10}, [1, 2, 3, 4, 5, 6]),
        ({"gt": 10}, [2, 3, 4, 5]),
        ({"ge": 10, "lt": 12}, [1, 2, 6]),
        ({"isin": [11, 20]}, [2, 4]),
    ],
)
def test_compare_int(value, expected_true_indices):
    obj = pd.Series([0, 10, 11, 12, 20, 21, 10, 2])
    expected = np.full(len(obj), False)
    expected[expected_true_indices] = True

    result = test_module.compare(obj, value)

    assert_array_equal(result, expected)


@pytest.mark.parametrize(
    "value, expected_true_indices",
    [
        ("s10", [1, 6]),
        ("s-1", []),
        (["s11", "s20"], [2, 4]),
        ({"eq": "s10"}, [1, 6]),
        ({"ne": "s10"}, [0, 2, 3, 4, 5, 7]),
        ({"le": "s10"}, [0, 1, 6]),
        ({"lt": "s10"}, [0]),
        ({"ge": "s10"}, [1, 2, 3, 4, 5, 6, 7]),
        ({"gt": "s10"}, [2, 3, 4, 5, 7]),
        ({"ge": "s10", "lt": "s12"}, [1, 2, 6]),
        ({"isin": ["s11", "s20"]}, [2, 4]),
        ({"regex": "10"}, [1, 6]),
        ({"regex": "^10"}, []),
        ({"regex": "s2"}, [4, 5, 7]),
        ({"regex": "^s2"}, [4, 5, 7]),
    ],
)
def test_compare_str(value, expected_true_indices):
    obj = pd.Series(["s0", "s10", "s11", "s12", "s20", "s21", "s10", "s2"])
    expected = np.full(len(obj), False)
    expected[expected_true_indices] = True

    result = test_module.compare(obj, value)

    assert_array_equal(result, expected)


def test_compare_unsupported_operator():
    obj = pd.Series([0, 10, 11, 12, 20, 21, 10])

    with pytest.raises(ValueError, match="Unsupported operator"):
        test_module.compare(obj, {"unknown": 10})


def test_compare_empty_filter():
    obj = pd.Series([0, 10, 11, 12, 20, 21, 10])

    with pytest.raises(ValueError, match="Empty filter"):
        test_module.compare(obj, {})


@pytest.mark.parametrize(
    "left, right, expected",
    [
        ({}, {}, True),
        ({}, {"key": 1}, False),
        ({"key": 1}, {}, True),
        ({"key": 1}, {"key": 1}, True),
        ({"key": 1}, {"key": [1, 2]}, True),
        ({"key": 1}, {"key": {"isin": [1, 2]}}, True),
        ({"key": 1}, {"key": 2}, False),
        ({"key": 1}, {"key": [2, 3]}, False),
        ({"key": 1}, {"key": {"isin": [2, 3]}}, False),
        ({"key1": 1, "key2": 2}, {"key1": 1}, True),
        ({"key1": 1}, {"key1": 1, "key2": 2}, False),
        ({"key": {"isin": [1, 2]}}, {"key": 1}, False),
        ({"key": {"ne": 3}}, {"key": {"ne": 3}}, True),
        ({"key": {"ne": 3}}, {"key": {"ne": 4}}, False),
        ({"key": {"gt": 3}}, {"key": {"gt": 2}}, True),
        ({"key": {"gt": 3}}, {"key": {"gt": 3}}, True),
        ({"key": {"gt": 3}}, {"key": {"gt": 4}}, False),
        ({"key": {"ge": 3}}, {"key": {"ge": 2}}, True),
        ({"key": {"ge": 3}}, {"key": {"ge": 3}}, True),
        ({"key": {"ge": 3}}, {"key": {"ge": 4}}, False),
        ({"key": {"lt": 3}}, {"key": {"lt": 2}}, False),
        ({"key": {"lt": 3}}, {"key": {"lt": 3}}, True),
        ({"key": {"lt": 3}}, {"key": {"lt": 4}}, True),
        ({"key": {"le": 3}}, {"key": {"le": 2}}, False),
        ({"key": {"le": 3}}, {"key": {"le": 3}}, True),
        ({"key": {"le": 3}}, {"key": {"le": 4}}, True),
        ({"key": {"le": 3, "ge": 1}}, {"key": {"le": 4}}, True),
        ({"key": {"le": 3, "ge": 1}}, {"key": {"le": 4, "ge": 0}}, True),
        ({"key": 1}, {"key": {"eq": 1}}, True),
        ({"key": 1}, {"key": {"eq": 1, "isin": [1, 2]}}, True),
        ({"key": 1}, {"key": {"eq": 1, "isin": [2, 3]}}, False),
    ],
)
def test_is_subfilter(left, right, expected):
    result = test_module.is_subfilter(left, right)
    assert result == expected


def test_smart_concat_series(series1):
    obj1 = series1.copy() + 1
    obj2 = series1.copy() + 2
    iterable = [obj1, obj2]
    expected = pd.Series(
        [1, 2, 3, 4, 2, 3, 4, 5],
        index=pd.MultiIndex.from_tuples(
            [("a", "c"), ("a", "d"), ("b", "c"), ("b", "d")] * 2, names=["i0", "i1"]
        ),
        name="values",
    )

    result = test_module.smart_concat(iterable)
    assert_series_equal(result, expected)

    expected = pd.Series(
        [1, 2, 3, 4, 2, 3, 4, 5],
        index=pd.MultiIndex.from_tuples(
            [
                ("e", "a", "c"),
                ("e", "a", "d"),
                ("e", "b", "c"),
                ("e", "b", "d"),
                ("f", "a", "c"),
                ("f", "a", "d"),
                ("f", "b", "c"),
                ("f", "b", "d"),
            ],
            names=["i2", "i0", "i1"],
        ),
        name="values",
    )
    result = test_module.smart_concat(iterable, keys=["e", "f"], names=["i2"])
    assert_series_equal(result, expected)

    result = test_module.smart_concat({"e": obj1, "f": obj2}, names=["i2"])
    assert_series_equal(result, expected)


def test_smart_concat_series_having_indexes_with_different_level_order(series1):
    # test that the result is consistent when the levels of the indexes are ordered differently.
    # plain pd.concat would blindly concatenate the indexes, ignoring the names of the levels.
    obj1 = series1.copy() + 1
    obj2 = series1.copy().reorder_levels(["i1", "i0"]) + 2
    iterable = [obj1, obj2]

    expected = pd.Series(
        [1, 2, 3, 4, 2, 3, 4, 5],
        index=pd.MultiIndex.from_tuples(
            [("a", "c"), ("a", "d"), ("b", "c"), ("b", "d")] * 2, names=["i0", "i1"]
        ),
        name="values",
    )
    result = test_module.smart_concat(iterable)
    assert_series_equal(result, expected)

    expected = pd.Series(
        [1, 2, 3, 4, 2, 3, 4, 5],
        index=pd.MultiIndex.from_tuples(
            [
                ("e", "a", "c"),
                ("e", "a", "d"),
                ("e", "b", "c"),
                ("e", "b", "d"),
                ("f", "a", "c"),
                ("f", "a", "d"),
                ("f", "b", "c"),
                ("f", "b", "d"),
            ],
            names=["i2", "i0", "i1"],
        ),
        name="values",
    )
    result = test_module.smart_concat(iterable, keys=["e", "f"], names=["i2"])
    assert_series_equal(result, expected)

    result = test_module.smart_concat({"e": obj1, "f": obj2}, names=["i2"])
    assert_series_equal(result, expected)


def test_smart_concat_series_having_indexes_with_different_level_name(series1):
    # test that an exception is raised when the levels of the indexes are different.
    # plain pd.concat would blindly concatenate the indexes, ignoring the names of the levels.
    obj1 = series1.copy() + 1
    obj2 = series1.copy() + 2
    obj2.index.set_names({"i0": "x0"}, inplace=True)
    iterable = [obj1, obj2]
    with pytest.raises(RuntimeError, match="Levels not found: i0"):
        test_module.smart_concat(iterable)
    with pytest.raises(RuntimeError, match="Levels not found: i0"):
        test_module.smart_concat({"e": obj1, "f": obj2})


def test_smart_concat_series_having_indexes_with_different_number_of_levels(series1):
    # test that an exception is raised when the levels of the indexes are different.
    # plain pd.concat would blindly concatenate the indexes, ignoring the names of the levels.
    obj1 = series1.copy() + 1
    obj2 = series1.copy() + 2
    obj2 = obj2.to_frame()
    obj2["i2"] = 999
    obj2 = obj2.set_index("i2", append=True)["values"]
    iterable = [obj1, obj2]
    match = re.escape("Length of order must be same as number of levels (3), got 2")
    with pytest.raises(RuntimeError, match=match):
        test_module.smart_concat(iterable)
    with pytest.raises(RuntimeError, match=match):
        test_module.smart_concat({"e": obj1, "f": obj2})


def test_smart_concat_some_empty_series(series1):
    obj1 = series1.astype(np.int32)
    obj2 = series1.astype(np.float32).iloc[:0]
    iterable = [obj1, obj2]
    result = test_module.smart_concat(iterable, skip_empty=True)
    assert_series_equal(result, obj1)


def test_smart_concat_all_empty_series(series1):
    obj1 = series1.astype(np.int32).iloc[:0]
    obj2 = series1.astype(np.int16).iloc[:0]
    iterable = [obj1, obj2]
    result = test_module.smart_concat(iterable, skip_empty=True)
    assert_series_equal(result, obj1)


def test_smart_concat_dataframes(dataframe1):
    obj1 = dataframe1.copy() + 1
    obj2 = dataframe1.copy() + 2
    iterable = [obj1, obj2]
    expected = pd.DataFrame(
        {"v0": [1, 2, 3, 4, 2, 3, 4, 5], "v1": [5, 6, 7, 8, 6, 7, 8, 9]},
        index=pd.MultiIndex.from_tuples(
            [("a", "c"), ("a", "d"), ("b", "c"), ("b", "d")] * 2, names=["i0", "i1"]
        ),
    )
    result = test_module.smart_concat(iterable)
    assert_frame_equal(result, expected)

    expected = pd.DataFrame(
        {"v0": [1, 2, 3, 4, 2, 3, 4, 5], "v1": [5, 6, 7, 8, 6, 7, 8, 9]},
        index=pd.MultiIndex.from_tuples(
            [
                ("e", "a", "c"),
                ("e", "a", "d"),
                ("e", "b", "c"),
                ("e", "b", "d"),
                ("f", "a", "c"),
                ("f", "a", "d"),
                ("f", "b", "c"),
                ("f", "b", "d"),
            ],
            names=["i2", "i0", "i1"],
        ),
    )
    result = test_module.smart_concat(iterable, keys=["e", "f"], names=["i2"])
    assert_frame_equal(result, expected)

    result = test_module.smart_concat({"e": obj1, "f": obj2}, names=["i2"])
    assert_frame_equal(result, expected)


def test_smart_concat_dataframes_having_indexes_with_different_level_order(dataframe1):
    # test that the result is consistent when the levels of the indexes are ordered differently.
    # plain pd.concat would blindly concatenate the indexes, ignoring the names of the levels.
    obj1 = dataframe1.copy() + 1
    obj2 = dataframe1.copy().reorder_levels(["i1", "i0"]) + 2
    iterable = [obj1, obj2]
    expected = pd.DataFrame(
        {"v0": [1, 2, 3, 4, 2, 3, 4, 5], "v1": [5, 6, 7, 8, 6, 7, 8, 9]},
        index=pd.MultiIndex.from_tuples(
            [("a", "c"), ("a", "d"), ("b", "c"), ("b", "d")] * 2, names=["i0", "i1"]
        ),
    )
    result = test_module.smart_concat(iterable)
    assert_frame_equal(result, expected)

    expected = pd.DataFrame(
        {"v0": [1, 2, 3, 4, 2, 3, 4, 5], "v1": [5, 6, 7, 8, 6, 7, 8, 9]},
        index=pd.MultiIndex.from_tuples(
            [
                ("e", "a", "c"),
                ("e", "a", "d"),
                ("e", "b", "c"),
                ("e", "b", "d"),
                ("f", "a", "c"),
                ("f", "a", "d"),
                ("f", "b", "c"),
                ("f", "b", "d"),
            ],
            names=["i2", "i0", "i1"],
        ),
    )
    result = test_module.smart_concat(iterable, keys=["e", "f"], names=["i2"])
    assert_frame_equal(result, expected)

    result = test_module.smart_concat({"e": obj1, "f": obj2}, names=["i2"])
    assert_frame_equal(result, expected)


def test_smart_concat_dataframes_having_indexes_with_different_level_name(dataframe1):
    # test that an exception is raised when the levels of the indexes are different.
    # plain pd.concat would blindly concatenate the indexes, ignoring the names of the levels.
    obj1 = dataframe1.copy() + 1
    obj2 = dataframe1.copy() + 2
    obj2.index.set_names({"i0": "x0"}, inplace=True)
    iterable = [obj1, obj2]
    with pytest.raises(RuntimeError, match="Levels not found: i0"):
        test_module.smart_concat(iterable)
    with pytest.raises(RuntimeError, match="Levels not found: i0"):
        test_module.smart_concat({"e": obj1, "f": obj2})


def test_smart_concat_dataframes_having_indexes_with_different_number_of_levels(dataframe1):
    # test that an exception is raised when the levels of the indexes are different.
    # plain pd.concat would blindly concatenate the indexes, ignoring the names of the levels.
    obj1 = dataframe1.copy() + 1
    obj2 = dataframe1.copy() + 2
    obj2["i2"] = 999
    obj2 = obj2.set_index("i2", append=True)
    iterable = [obj1, obj2]
    match = re.escape("Length of order must be same as number of levels (3), got 2")
    with pytest.raises(RuntimeError, match=match):
        test_module.smart_concat(iterable)
    with pytest.raises(RuntimeError, match=match):
        test_module.smart_concat({"e": obj1, "f": obj2})


def test_smart_concat_some_empty_dataframes(dataframe1):
    obj1 = dataframe1.astype({"v0": np.int32})
    obj2 = dataframe1.astype({"v0": np.float32}).iloc[:0]
    iterable = [obj1, obj2]
    result = test_module.smart_concat(iterable, skip_empty=True)
    assert_frame_equal(result, obj1)


def test_smart_concat_all_empty_dataframes(dataframe1):
    obj1 = dataframe1.astype({"v0": np.int32}).iloc[:0]
    obj2 = dataframe1.astype({"v0": np.int16}).iloc[:0]
    iterable = [obj1, obj2]
    result = test_module.smart_concat(iterable, skip_empty=True)
    assert_frame_equal(result, obj1)


@pytest.mark.parametrize(
    "iterables, expected",
    [
        ([[], []], 0),
        ([[], [1, 2, 3]], 0),
        ([[1, 2], [1, 2, 3]], 2),
        ([[1, 2, 3], [1, 2]], 2),
        ([[1, 2, 3], [1, 2, 3]], 3),
        ([[2, 3, 1], [1, 2, 3]], 0),
        ([[1, 4, 3], [1, 2, 3]], 1),
        ([(i for i in "abc"), ["a", "b", "c"]], 3),
        ([(i for i in "acc"), ["a", "b", "c"]], 1),
    ],
)
def test_longest_match_count(iterables, expected):
    result = test_module.longest_match_count(*iterables)
    assert result == expected


def test_cached_dataframe():
    df = pd.DataFrame(
        {
            "simulation_id": [0, 0, 0, 0, 1, 1, 1, 1],
            "circuit_id": [0, 0, 0, 0, 0, 0, 0, 0],
            "window": ["w0", "w0", "w1", "w1", "w0", "w0", "w1", "w1"],
            "trial": [0, 1, 0, 1, 0, 1, 0, 1],
        }
    )
    query = {"simulation_id": 1, "circuit_id": 0, "window": "w0"}
    expected_df = pd.DataFrame(
        {
            "simulation_id": [1, 1],
            "circuit_id": [0, 0],
            "window": ["w0", "w0"],
            "trial": [0, 1],
        },
        index=[4, 5],
    )
    expected_stack = [
        test_module.CachedItem(
            df=pd.DataFrame(
                {
                    "simulation_id": [1, 1, 1, 1],
                    "circuit_id": [0, 0, 0, 0],
                    "window": ["w0", "w0", "w1", "w1"],
                    "trial": [0, 1, 0, 1],
                },
                index=[4, 5, 6, 7],
            ),
            key="simulation_id",
            value=1,
        ),
        test_module.CachedItem(
            df=pd.DataFrame(
                {
                    "simulation_id": [1, 1, 1, 1],
                    "circuit_id": [0, 0, 0, 0],
                    "window": ["w0", "w0", "w1", "w1"],
                    "trial": [0, 1, 0, 1],
                },
                index=[4, 5, 6, 7],
            ),
            key="circuit_id",
            value=0,
        ),
        test_module.CachedItem(
            df=pd.DataFrame(
                {
                    "simulation_id": [1, 1],
                    "circuit_id": [0, 0],
                    "window": ["w0", "w0"],
                    "trial": [0, 1],
                },
                index=[4, 5],
            ),
            key="window",
            value="w0",
        ),
    ]
    cache = test_module.CachedDataFrame(df=df)

    result = cache.query(query=query)
    assert_frame_equal(result, expected_df)
    assert cache._stack == expected_stack
    assert cache._matched == 0

    query["trial"] = 1

    result = cache.query(query=query)
    assert_frame_equal(result, expected_df.iloc[[1]])
    assert cache._stack == expected_stack + [
        test_module.CachedItem(
            df=pd.DataFrame(
                {
                    "simulation_id": [1],
                    "circuit_id": [0],
                    "window": ["w0"],
                    "trial": [1],
                },
                index=[5],
            ),
            key="trial",
            value=1,
        )
    ]
    assert cache._matched == len(expected_stack)

    query["trial"] = 0

    result = cache.query(query=query)
    assert_frame_equal(result, expected_df.iloc[[0]])
    assert cache._stack == expected_stack + [
        test_module.CachedItem(
            df=pd.DataFrame(
                {
                    "simulation_id": [1],
                    "circuit_id": [0],
                    "window": ["w0"],
                    "trial": [0],
                },
                index=[4],
            ),
            key="trial",
            value=0,
        )
    ]
    assert cache._matched == len(expected_stack)

    del query["trial"]

    result = cache.query(query=query)
    assert_frame_equal(result, expected_df)
    assert cache._stack == expected_stack
    assert cache._matched == len(expected_stack)



"""Core utilities."""
import operator
from collections.abc import Callable, Mapping
from copy import deepcopy
from dataclasses import dataclass
from itertools import chain
from operator import attrgetter
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from blueetl_core.logging import L

COMPARISON_OPERATORS = {
    "eq": attrgetter("__eq__"),
    "ne": attrgetter("__ne__"),
    "le": attrgetter("__le__"),
    "lt": attrgetter("__lt__"),
    "ge": attrgetter("__ge__"),
    "gt": attrgetter("__gt__"),
    "isin": attrgetter("isin"),
    "regex": attrgetter("str.contains"),
}


def ensure_list(x: Any) -> list:
    """Always return a list from the given argument."""
    if isinstance(x, list):
        return x
    if isinstance(x, tuple):
        return list(x)
    return [x]


def _and_or_mask(
    query_list: list[dict[str, Any]],
    filter_func: Callable[[dict[str, Any]], list[np.ndarray]],
) -> Optional[np.ndarray]:
    """Return the and/or mask obtained after processing each dictionary in query_list.

    Args:
        query_list: list of query dictionaries, that should be processed and OR-ed together.
        filter_func: callable accepting a query dict and returning a list of masks,
            that will be AND-ed together.

    Returns:
        The resulting mask, or None if no queries have been provided.
    """
    or_masks = []
    for query in query_list:
        if query:
            and_masks = filter_func(query)
            # minor optimization: check if len == 1
            or_masks.append(and_masks[0] if len(and_masks) == 1 else np.all(and_masks, axis=0))
    if not or_masks:
        # no filters
        return None
    # minor optimization: check if len == 1
    return or_masks[0] if len(or_masks) == 1 else np.any(or_masks, axis=0)


def query_frame(df: pd.DataFrame, query_list: list[dict[str, Any]]) -> pd.DataFrame:
    """Given a query dictionary, return the DataFrame filtered by columns and index."""

    def _filter_func(query: dict[str, Any]) -> list[np.ndarray]:
        # dictionary with query keys split into columns and index
        q: dict[str, Any] = {"columns": {}, "index": {}}
        for key, value in query.items():
            q[mapping[key]][key] = value
        # filter by columns and index
        return list(
            chain(
                (compare(df[key], val) for key, val in q["columns"].items()),
                (compare(df.index.get_level_values(key), val) for key, val in q["index"].items()),
            )
        )

    # map each name to columns or index;
    # if the same key is present in both columns and index, use columns
    mapping = {
        **{k: "index" for k in df.index.names if k is not None},
        **{k: "columns" for k in df.columns if k is not None},
    }
    mask = _and_or_mask(query_list, _filter_func)
    return df.loc[mask] if mask is not None else df


def query_series(series: pd.Series, query_list: list[dict[str, Any]]) -> pd.Series:
    """Given a query dictionary, return the Series filtered by index."""

    def _filter_func(query: dict[str, Any]) -> list[np.ndarray]:
        # filter by index
        return list(compare(series.index.get_level_values(key), val) for key, val in query.items())

    mask = _and_or_mask(query_list, _filter_func)
    return series.loc[mask] if mask is not None else series


def compare(obj: Union[pd.Series, pd.Index], value: Any) -> np.ndarray:
    """Return the result of the comparison between obj and value.

    Args:
        obj: Series, or Index.
        value: value used for comparison.
            - if scalar, use equality
            - if list-like, use isin
            - if dict, any supported operators can be specified, and they will be AND-ed together

    Examples:
        >>> df = pd.DataFrame({"gid": [0, 2, 3, 7, 8]})
        >>> compare(df["gid"], 3)
            array([False, False,  True, False, False])
        >>> compare(df["gid"], [3, 5, 8])
            array([False, False,  True, False,  True])
        >>> compare(df["gid"], {"ge": 3, "lt": 8})
            array([False, False,  True,  True, False])

    """
    if isinstance(value, dict):
        if not value:
            raise ValueError("Empty filter")
        if unsupported := [op for op in value if op not in COMPARISON_OPERATORS]:
            raise ValueError(f"Unsupported operator(s): {unsupported}")
        masks = [COMPARISON_OPERATORS[op](obj)(v) for op, v in value.items()]
        return masks[0] if len(masks) == 1 else np.all(masks, axis=0)
    if is_list_like(value):
        return np.asarray(obj.isin(value))
    # more efficient than using isin with a list of one element
    return np.asarray(obj == value)


def is_subfilter(left: dict, right: dict) -> bool:
    """Return True if ``left`` is a subfilter of ``right``, False otherwise.

    ``left`` is a subfilter of ``right`` if it's equal or more specific.

    Examples:
        >>> print(is_subfilter({}, {}))
        True
        >>> print(is_subfilter({}, {"key": 1}))
        False
        >>> print(is_subfilter({"key": 1}, {}))
        True
        >>> print(is_subfilter({"key": 1}, {"key": 1}))
        True
        >>> print(is_subfilter({"key": 1}, {"key": [1, 2]}))
        True
        >>> print(is_subfilter({"key": 1}, {"key": {"isin": [1, 2]}}))
        True
        >>> print(is_subfilter({"key": 1}, {"key": 2}))
        False
        >>> print(is_subfilter({"key": 1}, {"key": [2, 3]}))
        False
        >>> print(is_subfilter({"key": 1}, {"key": {"isin": [2, 3]}}))
        False
        >>> print(is_subfilter({"key1": 1, "key2": 2}, {"key1": 1}))
        True
        >>> print(is_subfilter({"key1": 1}, {"key1": 1, "key2": 2}))
        False
    """

    def _to_dict(obj) -> dict:
        """Return a normalized filter, i.e. a dict where "eq" is replaced by "isin"."""
        obj = deepcopy(obj)
        if isinstance(obj, dict):
            if "eq" in obj:
                # convert "eq" to "isin", and set "isin" to the new value,
                # or to an empty list if "eq" and "isin" are incompatible
                value = obj.pop("eq")
                obj["isin"] = [value] if "isin" not in obj or value in obj["isin"] else []
            return obj
        if isinstance(obj, list):
            return {"isin": obj}
        # any other type of object is considered for equality with "isin"
        return {"isin": [obj]}

    def _is_subdict(d1: dict, d2: dict) -> bool:
        """Return True if d1 is a subdict of d2."""
        # mapping operator -> operation
        operators = {
            "ne": operator.eq,
            "le": operator.le,
            "lt": operator.le,
            "ge": operator.ge,
            "gt": operator.ge,
            "isin": lambda a, b: set(a).issubset(b),
        }
        assert set(operators).issuperset(d1), "Invalid keys in d1"
        assert set(operators).issuperset(d2), "Invalid keys in d2"
        unmatched_keys = set()
        # for each operator in the operators mapping,
        # if the operator is present in d2 but not in d1,
        # or if the given operation is not satisfied,
        # then d1 cannot be a subdict of d2
        for op, operation in operators.items():
            if op in d2 and (op not in d1 or not operation(d1[op], d2[op])):
                unmatched_keys.add(op)
        L.debug("unmatched keys: %s", sorted(unmatched_keys))
        return len(unmatched_keys) == 0

    for key in right:
        if key not in left:
            return False
        dict_left = _to_dict(left[key])
        dict_right = _to_dict(right[key])
        if not _is_subdict(dict_left, dict_right):
            return False
    return True


def smart_concat(iterable, *, keys=None, copy=False, skip_empty=True, **kwargs):
    """Build and return a Series or a Dataframe from an iterable of objects with the same index.

    This is similar to ``pd.concat``, but the result is consistent even when the levels of the
    indexes are ordered differently, while ``pd.concat`` would blindly concatenate the indexes,
    ignoring and removing the names of the levels.

    Moreover, it uses ``copy=False`` by default, that's safe only if the original data isn't going
    to change, but it's more efficient, especially when concatenating a single item.

    Args:
        iterable: iterable or mapping of Series or DataFrames.
            All the objects must be of the same type, and they must have the same index,
            or an exception is raised.
        keys: passed to pd.concat. If multiple levels passed, should contain tuples. Construct
            hierarchical index using the passed keys as the outermost level.
        copy: passed to pd.concat. If the original data can be used without making a copy, then
            it can be set to False.
        skip_empty: if True, empty objects are skipped, unless they are all empty. If False, they
            are all passed to pd.concat, and the result may depend on the Pandas version.
            Note that in the latter case, you may see a FutureWarning with Pandas 2:

                FutureWarning: The behavior of DataFrame concatenation with empty or all-NA entries
                is deprecated. In a future version, this will no longer exclude empty or all-NA
                columns when determining the result dtypes. To retain the old behavior, exclude the
                relevant entries before the concat operation.
        kwargs: other keyword arguments to be passed to pd.concat

    Returns:
        (pd.Series|pd.DataFrame) result of the concatenation, same type of the input elements.

    Examples:
        >>> idx1 = pd.MultiIndex.from_tuples([(10, 11), (20, 21)], names=["i1", "i2"])
        >>> idx2 = pd.MultiIndex.from_tuples([(11, 10), (31, 30)], names=["i2", "i1"])
        >>> df1 = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=idx1)
        >>> df2 = pd.DataFrame({"A": [5, 6], "B": [7, 8]}, index=idx2)
        >>> pd.concat([df1, df2])  # index levels are lost
               A  B
        10 11  1  3
        20 21  2  4
        11 10  5  7
        31 30  6  8
        >>> smart_concat([df1, df2])  # index levels are preserved
               A  B
        i1 i2
        10 11  1  3
        20 21  2  4
        10 11  5  7
        30 31  6  8
        >>> pd.concat([df1, df2], axis=1)  # index levels are lost
                 A    B    A    B
        10 11  1.0  3.0  NaN  NaN
        20 21  2.0  4.0  NaN  NaN
        11 10  NaN  NaN  5.0  7.0
        31 30  NaN  NaN  6.0  8.0
        >>> smart_concat([df1, df2], axis=1)  # index levels are preserved
                 A    B    A    B
        i1 i2
        10 11  1.0  3.0  5.0  7.0
        20 21  2.0  4.0  NaN  NaN
        30 31  NaN  NaN  6.0  8.0
    """

    def _reorder_levels(obj, order):
        # wrap reorder_levels to raise an explicit error when needed
        if len(order) != obj.index.nlevels:
            # reorder_levels would raise an AssertionError
            raise RuntimeError(
                f"Length of order must be same as number of "
                f"levels ({obj.index.nlevels}), got {len(order)}"
            )
        if diff := set(order).difference(obj.index.names):
            # reorder_levels would raise a KeyError
            raise RuntimeError(f"Levels not found: {''.join(diff)}")
        return obj.reorder_levels(order)

    def _ordered(obj):
        nonlocal order
        if order is None:
            order = obj.index.names
        return obj if order == obj.index.names else _reorder_levels(obj, order)

    order = None
    if isinstance(iterable, Mapping):
        mapping = iterable
        keys = keys if keys is not None else mapping.keys()
        iterable = (mapping[key] for key in keys)
    objects = [_ordered(obj) for obj in iterable]
    if skip_empty and not all(obj.empty for obj in objects):
        objects = [obj for obj in objects if not obj.empty]
    return pd.concat(objects, keys=keys, copy=copy, **kwargs)


def concat_tuples(iterable, *args, **kwargs):
    """Build and return a Series from an iterable of tuples (value, conditions).

    Args:
        iterable: iterable of tuples (value, conditions), where

            - value is a single value that will be added to the Series
            - conditions is a dict containing the conditions to be used for the MultiIndex.
              The keys of the conditions must be the same for each tuple of the iterable,
              or an exception is raised.

        args: positional arguments to be passed to pd.concat
        kwargs: key arguments to be passed to pd.concat

    Returns:
        (pd.Series) result of the concatenation.
    """

    def _index(conditions):
        arrays = [[v] for v in conditions.values()]
        names = list(conditions)
        return pd.MultiIndex.from_arrays(arrays, names=names)

    iterable = (pd.Series([data], index=_index(conditions)) for data, conditions in iterable)
    return smart_concat(iterable, *args, **kwargs)


def longest_match_count(iter1, iter2) -> int:
    """Return the number of matching elements from the beginning of the given iterables."""
    count = 0
    for i1, i2 in zip(iter1, iter2):
        if i1 != i2:
            break
        count += 1
    return count


@dataclass
class CachedItem:
    """Item of CachedDataFrame."""

    df: pd.DataFrame
    key: str
    value: Any

    def __eq__(self, other: object) -> bool:
        """Return True if the objects are considered equal, False otherwise."""
        if not isinstance(other, CachedItem):
            return NotImplemented
        return self.key == other.key and self.value == other.value and self.df.equals(other.df)


class CachedDataFrame:
    """DataFrame wrapper to cache partial queries."""

    def __init__(self, df: pd.DataFrame) -> None:
        """Initialize the object with the base DataFrame.

        The internal stack will contain CachedItems, each one containing a DataFrame filtered by
        the corresponding key and value, and by all the previous keys and values in the stack.

        Examples:
            .. code-block:: python

                self._stack = [
                    CachedItem(df=df0, key="simulation_id", value=1),
                    CachedItem(df=df1, key="circuit_id", value=0),
                    CachedItem(df=df2, key="window", value="w1"),
                    CachedItem(df=df3, key="trial", value=0),
                ]

            where:

            - ``df0`` is ``self._df`` filtered by ``simulation_id=1``
            - ``df1`` is ``df0`` filtered by ``circuit_id=0``
            - ``df2`` is ``df1`` filtered by ``window="w1"``
            - ``df3`` is ``df2`` filtered by ``trial=0``

        """
        self._df = df
        self._valid_keys = {*df.columns, *(key for key in df.index.names if key)}
        self._stack: list[CachedItem] = []
        self._matched = 0  # for test and debug

    def _longest_keys_count(self, keys) -> int:
        return longest_match_count((item.key for item in self._stack), keys)

    def _longest_values_count(self, values) -> int:
        return longest_match_count((item.value for item in self._stack), values)

    def query(self, query: dict[str, Any], ignore_unknown_keys: bool = False) -> pd.DataFrame:
        """Return the DataFrame filtered by query, using cached DataFrames if possible.

        - The order of the keys in the query dict is important.
        - The cache is reused only when the keys and their order are the same.
        - The cache is reused also when only some keys and their values match.

        Args:
            query: dict to be passed to ``etl.q``.
            ignore_unknown_keys: if True, ignore keys specified in the query but not present in the
            DataFrame columns or in the index level names. If False, unknown keys raise an error.

        """
        if ignore_unknown_keys:
            query = {key: value for key, value in query.items() if key in self._valid_keys}
        query_keys = tuple(query.keys())
        query_values = tuple(query.values())
        # find the cached dataframe with the longest key
        self._matched = min(
            self._longest_keys_count(query_keys),
            self._longest_values_count(query_values),
        )
        self._stack = self._stack[: self._matched]
        df = self._stack[-1].df if self._stack else self._df
        # update the cache for every partial key, if needed
        while len(self._stack) < len(query):
            col = query_keys[len(self._stack)]
            val = query_values[len(self._stack)]
            df = df.etl.q({col: val})
            self._stack.append(CachedItem(df=df, key=col, value=val))
        return df



from typing import Iterator

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal, assert_equal
from pandas.testing import assert_index_equal, assert_series_equal

from blueetl_core.etl import ETLSeriesAccessor


def test_etl_instance(series1):
    obj = series1
    result = obj.etl
    assert isinstance(result, ETLSeriesAccessor)


def test_conditions(series1):
    obj = series1
    result = obj.etl.conditions()
    assert_array_equal(result, ["i0", "i1"])


@pytest.mark.parametrize(
    "conditions, expected",
    [
        pytest.param("i0", ["i1"], id="single"),
        pytest.param(["i0"], ["i1"], id="list"),
        pytest.param([], ["i0", "i1"], id="empty"),
        pytest.param(["i0", "i1"], [], id="all"),
        pytest.param(["iX"], ["i0", "i1"], id="nonexistent"),
    ],
)
def test_complementary_conditions(series1, conditions, expected):
    obj = series1
    result = obj.etl.complementary_conditions(conditions)
    assert_array_equal(result, expected)


@pytest.mark.parametrize(
    "conditions, expected",
    [
        pytest.param(None, [pd.Index(["a", "b"], name="i0"), pd.Index(["c", "d"], name="i1")]),
        pytest.param(["i0"], [pd.Index(["a", "b"], name="i0")], id="one"),
    ],
)
def test_labels(series1, conditions, expected):
    obj = series1
    result = obj.etl.labels(conditions)
    assert len(result) == len(expected)
    for actual_index, expected_index in zip(result, expected):
        assert_index_equal(actual_index, expected_index)


def test_labels_of(series1):
    obj = series1
    result = obj.etl.labels_of("i0")
    assert_array_equal(result, ["a", "b"])


def test_labels_of_nonexistent_condition(series1):
    obj = series1
    with pytest.raises(KeyError, match="Level iX not found"):
        obj.etl.labels_of("iX")


@pytest.mark.parametrize(
    "conditions",
    [
        pytest.param("i0", id="str"),
        pytest.param(["i0"], id="list"),
    ],
)
def test_remove_conditions(series1, conditions):
    obj = series1
    result = obj.etl.remove_conditions(conditions)
    assert_series_equal(
        result,
        pd.Series(
            [0, 1, 2, 3],
            index=pd.Index(["c", "d", "c", "d"], name="i1"),
            name="values",
        ),
    )


@pytest.mark.parametrize(
    "conditions",
    [
        pytest.param("i1", id="str"),
        pytest.param(["i1"], id="list"),
    ],
)
def test_keep_conditions(series1, conditions):
    obj = series1
    result = obj.etl.keep_conditions(conditions)
    assert_series_equal(
        result,
        pd.Series(
            [0, 1, 2, 3],
            index=pd.Index(["c", "d", "c", "d"], name="i1"),
            name="values",
        ),
    )


@pytest.mark.parametrize(
    "conditions, values",
    [
        pytest.param("i2", "e", id="str"),
        pytest.param(["i2"], ["e"], id="list"),
    ],
)
def test_add_conditions(series1, conditions, values):
    obj = series1
    result = obj.etl.add_conditions(conditions, values)
    assert_series_equal(
        result,
        pd.Series(
            [0, 1, 2, 3],
            index=pd.MultiIndex.from_tuples(
                [("e", "a", "c"), ("e", "a", "d"), ("e", "b", "c"), ("e", "b", "d")],
                names=["i2", "i0", "i1"],
            ),
            name="values",
        ),
    )


@pytest.mark.parametrize(
    "conditions, values, dtypes",
    [
        pytest.param("i2", 123, float, id="str"),
        pytest.param(["i2"], [123], [float], id="list"),
    ],
)
def test_add_conditions_with_dtypes(series1, conditions, values, dtypes):
    obj = series1
    result = obj.etl.add_conditions(conditions, values, dtypes=dtypes)
    assert_series_equal(
        result,
        pd.Series(
            [0, 1, 2, 3],
            index=pd.MultiIndex.from_tuples(
                [(123.0, "a", "c"), (123.0, "a", "d"), (123.0, "b", "c"), (123.0, "b", "d")],
                names=["i2", "i0", "i1"],
            ),
            name="values",
        ),
    )
    assert_equal(result.index.dtypes.to_numpy(), [float, object, object])


@pytest.mark.parametrize(
    "inner, drop, expected_conditions",
    [
        (False, False, ["i2", "i0", "i1"]),
        (True, False, ["i0", "i1", "i2"]),
        (False, True, ["i2"]),
        (True, True, ["i2"]),
    ],
)
def test_add_conditions_inner_drop(series1, inner, drop, expected_conditions):
    obj = series1
    conditions = ["i2"]
    values = ["e"]
    result = obj.etl.add_conditions(conditions, values, inner=inner, drop=drop)
    assert result.index.names == expected_conditions


def test_select_with_drop_level(series1):
    obj = series1
    result = obj.etl.select(i1="c", drop_level=True)
    assert_series_equal(
        result,
        pd.Series(
            [0, 2],
            index=pd.Index(["a", "b"], name="i0"),
            name="values",
        ),
    )


def test_select_without_drop_level(series1):
    obj = series1
    result = obj.etl.select(i1="c", drop_level=False)
    assert_series_equal(
        result,
        pd.Series(
            [0, 2],
            index=pd.MultiIndex.from_tuples([("a", "c"), ("b", "c")], names=["i0", "i1"]),
            name="values",
        ),
    )


def test_select_without_args(series1):
    obj = series1
    result = obj.etl.select()
    assert_series_equal(result, obj)


def test_unpool_from_series_of_scalars(series1):
    def func(x):
        return pd.Series([x, x + 20], index=pd.Index(["x0", "x2"], name="func"))

    obj = series1
    result = obj.etl.unpool(func)
    assert_series_equal(
        result,
        pd.Series(
            [0, 20, 1, 21, 2, 22, 3, 23],
            index=pd.MultiIndex.from_tuples(
                [
                    ("a", "c", "x0"),
                    ("a", "c", "x2"),
                    ("a", "d", "x0"),
                    ("a", "d", "x2"),
                    ("b", "c", "x0"),
                    ("b", "c", "x2"),
                    ("b", "d", "x0"),
                    ("b", "d", "x2"),
                ],
                names=["i0", "i1", "func"],
            ),
            name=None,
        ),
    )


def test_unpool_from_series_of_lists(series1):
    def func(x):
        return pd.Series([x[0], x[2]], index=pd.Index(["x0", "x2"], name="func"))

    obj = series1
    # transform each element to a list of integers
    obj = obj.transform(lambda x: [x, x + 10, x + 20])
    result = obj.etl.unpool(func)
    assert_series_equal(
        result,
        pd.Series(
            [0, 20, 1, 21, 2, 22, 3, 23],
            index=pd.MultiIndex.from_tuples(
                [
                    ("a", "c", "x0"),
                    ("a", "c", "x2"),
                    ("a", "d", "x0"),
                    ("a", "d", "x2"),
                    ("b", "c", "x0"),
                    ("b", "c", "x2"),
                    ("b", "d", "x0"),
                    ("b", "d", "x2"),
                ],
                names=["i0", "i1", "func"],
            ),
            name=None,
        ),
    )


def test_pool_with_func_returning_scalar(series1):
    def func(x):
        return sum(x)

    obj = series1
    result = obj.etl.pool(["i1"], func)
    assert_series_equal(
        result,
        pd.Series(
            [1, 5],
            index=pd.Index(["a", "b"], name="i0"),
            name="values",
        ),
    )


def test_pool_with_func_returning_series(series1):
    def func(x):
        return pd.Series(
            [sum(x[:, "d"]), sum(x[:, "c"])], index=pd.Index(["x0", "x1"], name="func")
        )

    obj = series1
    result = obj.etl.pool(["i1"], func)
    assert_series_equal(
        result,
        pd.Series(
            [1, 0, 3, 2],
            index=pd.MultiIndex.from_tuples(
                [
                    ("a", "x0"),
                    ("a", "x1"),
                    ("b", "x0"),
                    ("b", "x1"),
                ],
                names=["i0", "func"],
            ),
            name="values",
        ),
    )


@pytest.mark.skip(reason="experimental")
def test_remodel(series1):
    def func(x):
        for index, value in x.etl.iter():
            yield value + 1000, {**index._asdict(), "i2": "e"}
            yield value + 2000, {**index._asdict(), "i2": "f"}

    obj = series1
    result = obj.etl.remodel(func)
    assert_series_equal(
        result,
        pd.Series(
            [1000, 2000, 1001, 2001, 1002, 2002, 1003, 2003],
            index=pd.MultiIndex.from_tuples(
                [
                    ("a", "c", "e"),
                    ("a", "c", "f"),
                    ("a", "d", "e"),
                    ("a", "d", "f"),
                    ("b", "c", "e"),
                    ("b", "c", "f"),
                    ("b", "d", "e"),
                    ("b", "d", "f"),
                ],
                names=["i0", "i1", "i2"],
            ),
        ),
    )


@pytest.mark.parametrize("as_type", ["dict", "list", "kwargs"])
@pytest.mark.parametrize(
    "query, expected",
    [
        (
            {"i0": "a"},
            pd.Series(
                [0, 1],
                index=pd.MultiIndex.from_tuples([("a", "c"), ("a", "d")], names=["i0", "i1"]),
                name="values",
            ),
        ),
        (
            {"i0": "a", "i1": "d"},
            pd.Series(
                [1],
                index=pd.MultiIndex.from_tuples([("a", "d")], names=["i0", "i1"]),
                name="values",
            ),
        ),
        (
            {"i0": "a", "i1": "x"},
            pd.Series(
                [],
                index=pd.MultiIndex.from_tuples([], names=["i0", "i1"]),
                name="values",
                dtype=np.int64,
            ),
        ),
        (
            {"i0": ["a", "b"], "i1": "c"},
            pd.Series(
                [0, 2],
                index=pd.MultiIndex.from_tuples([("a", "c"), ("b", "c")], names=["i0", "i1"]),
                name="values",
            ),
        ),
        (
            {"i0": {"ge": "b", "lt": "z"}, "i1": {"isin": ["d", "x"]}},
            pd.Series(
                [3],
                index=pd.MultiIndex.from_tuples([("b", "d")], names=["i0", "i1"]),
                name="values",
            ),
        ),
    ],
)
def test_query(series1, query, expected, as_type):
    obj = series1
    assert isinstance(query, dict)

    if as_type == "dict":
        result = obj.etl.q(query)
    elif as_type == "list":
        result = obj.etl.q([query])
    elif as_type == "kwargs":
        result = obj.etl.q(**query)
    else:
        raise ValueError(f"Invalid as_type: {as_type}")

    # check_index_type=False needed to avoid Attribute "inferred_type" are different
    assert_series_equal(result, expected, check_index_type=len(expected) != 0)


def test_query_with_multiple_dicts(series1):
    obj = series1
    query = [
        {"i0": "a"},
        {"i0": "b", "i1": "d"},
    ]
    expected = pd.Series(
        [0, 1, 3],
        index=pd.MultiIndex.from_tuples([("a", "c"), ("a", "d"), ("b", "d")], names=["i0", "i1"]),
        name="values",
    )

    result = obj.etl.q(query)

    assert_series_equal(result, expected)


def test_query_with_invalid_params(series1):
    obj = series1
    with pytest.raises(ValueError, match="Query and params cannot be specified together"):
        obj.etl.q({"i0": "a"}, i1="b")


def test_one(series1):
    obj = series1
    result = obj.etl.one(i0="b", i1="d")
    assert result == 3


def test_one_raises_when_multiple_results(series1):
    obj = series1
    with pytest.raises(RuntimeError, match="The query returned 2 records instead of 1."):
        obj.etl.one(i0="b")


def test_one_raises_when_no_results(series1):
    obj = series1
    with pytest.raises(RuntimeError, match="The query returned 0 records instead of 1."):
        obj.etl.one(i0="x")


def test_first(series1):
    obj = series1
    result = obj.etl.first(i0="b")
    assert result == 2


def test_first_raises_when_no_results(series1):
    obj = series1
    with pytest.raises(RuntimeError, match="The query returned 0 records."):
        obj.etl.first(i0="x")


def test_iter(series1):
    obj = series1
    it = obj.etl.iter()
    assert isinstance(it, Iterator)
    index, value = next(it)
    assert isinstance(index, tuple)
    assert index == ("a", "c")
    assert value == 0
    assert index.i0 == "a"
    assert index.i1 == "c"


def test_iterdict(series1):
    obj = series1
    it = obj.etl.iterdict()
    assert isinstance(it, Iterator)
    index, value = next(it)
    assert isinstance(index, dict)
    assert index == {"i0": "a", "i1": "c"}
    assert value == 0



"""BlueETL core constants."""

# Constants that can be set as environment variables to modify the behaviour of parallelization.

# If more than 10, all iterations are printed to stderr. Above 50, the output is sent to stdout.
# If not overridden, the value depends on logging: 0 if loglevel >= logging.WARNING else 10
BLUEETL_JOBLIB_VERBOSE = "BLUEETL_JOBLIB_VERBOSE"
# Number of concurrent jobs. If not overridden, it uses by default: os.cpu_count() // 2
BLUEETL_JOBLIB_JOBS = "BLUEETL_JOBLIB_JOBS"
# JobLib backend (loky, multiprocessing, threading). If not overridden, it uses by default: loky
BLUEETL_JOBLIB_BACKEND = "BLUEETL_JOBLIB_BACKEND"
# Subprocess logging level. If empty or not defined, fallback to the default level WARNING.
BLUEETL_SUBPROCESS_LOGGING_LEVEL = "BLUEETL_SUBPROCESS_LOGGING_LEVEL"



from typing import Iterator

from blueetl_core.etl import ETLIndexAccessor


def test_etl_instance(index1):
    obj = index1
    result = obj.etl
    assert isinstance(result, ETLIndexAccessor)


def test_iter(index1):
    obj = index1
    it = obj.etl.iter()
    assert isinstance(it, Iterator)
    item = next(it)
    assert isinstance(item, tuple)
    assert item == ("a", "c")
    assert item.i0 == "a"
    assert item.i1 == "c"


def test_iter_without_names(index2):
    obj = index2
    it = obj.etl.iter()
    assert isinstance(it, Iterator)
    item = next(it)
    assert isinstance(item, tuple)
    assert item == ("a", "c")
    assert item.ilevel_0 == "a"
    assert item.ilevel_1 == "c"


def test_iterdict(index1):
    obj = index1
    it = obj.etl.iterdict()
    assert isinstance(it, Iterator)
    item = next(it)
    assert isinstance(item, dict)
    assert item == {"i0": "a", "i1": "c"}


def test_iterdict_without_names(index2):
    obj = index2
    it = obj.etl.iterdict()
    assert isinstance(it, Iterator)
    item = next(it)
    assert isinstance(item, dict)
    assert item == {"ilevel_0": "a", "ilevel_1": "c"}



"""Pandas accessors."""
from abc import ABC, abstractmethod
from collections import namedtuple
from collections.abc import Iterator
from functools import partial
from typing import Any, Callable, Generic, NamedTuple, Optional, TypeVar, Union

import pandas as pd
from packaging.version import Version
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy

from blueetl_core.logging import L
from blueetl_core.parallel import Task, run_parallel
from blueetl_core.utils import ensure_list, query_frame, query_series, smart_concat

# Naming conventions
#  level: number, or name of the level in the MultiIndex
#  condition: name of the level (e.g. seed), similar to level, but it cannot be numeric
#  labels: possible values for a specific level of the index

PandasT = TypeVar("PandasT", pd.Series, pd.DataFrame)
PandasGroupByT = TypeVar("PandasGroupByT", SeriesGroupBy, DataFrameGroupBy)


class ETLBaseAccessor(ABC, Generic[PandasT, PandasGroupByT]):
    """Accessor with methods common to Series and DataFrame."""

    def __init__(self, pandas_obj: PandasT) -> None:
        """Initialize the accessor."""
        self._validate(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate(obj: PandasT) -> None:
        """Validate the wrapped object."""
        # assert all(obj.index.names), "All the index levels must have a name"
        # assert len(set(obj.index.names)) == obj.index.nlevels, "The level names must be unique"

    def conditions(self) -> list[str]:
        """Names for each of the index levels."""
        return self._obj.index.names

    def complementary_conditions(self, conditions: Union[str, list[str]]) -> list[str]:
        """Return the difference between the object conditions and the specified conditions.

        Args:
            conditions: single condition or list of conditions used to calculate the difference.
        """
        conditions = ensure_list(conditions)
        # TODO: raise a KeyError if any condition is not found in self.conditions?
        return self._obj.index.names.difference(conditions)

    def labels(self, conditions: Optional[list[str]] = None) -> list[pd.Index]:
        """Unique labels for each level, or for the specified levels.

        Args:
            conditions: list of condition names, or None to consider all the levels.

        Returns:
            list of indexes with unique labels, one for each level.
        """
        conditions = conditions or self.conditions()
        return [self.labels_of(condition) for condition in conditions]

    def labels_of(self, condition: str) -> pd.Index:
        """Unique labels for a specific level in the index.

        Args:
            condition: condition name.

        Returns:
            indexes with unique labels.
        """
        return self._obj.index.unique(condition)

    def remove_conditions(self, conditions: Union[str, list[str]]) -> PandasT:
        """Remove one or more conditions.

        Args:
            conditions: single condition or list of conditions to remove.

        Returns:
            resulting Series or DataFrame.
        """
        return self._obj.droplevel(conditions, axis=0)

    def keep_conditions(self, conditions: Union[str, list[str]]) -> PandasT:
        """Remove the conditions not specified.

        Args:
            conditions: single condition or list of conditions to keep.

        Returns:
            resulting Series or DataFrame.
        """
        return self._obj.droplevel(self.complementary_conditions(conditions), axis=0)

    def add_conditions(
        self,
        conditions: Union[str, list[str]],
        values: Any,
        inner: bool = False,
        drop: bool = False,
        dtypes: Any = None,
    ) -> PandasT:
        """Add one or multiple conditions into the outermost or innermost position.

        Args:
            conditions: single conditions or list of conditions to be added.
            values: single value or list of values, one for each condition.
            inner (bool): if True, add the conditions in the innermost position.
            drop (bool): if True, drop the existing conditions.
            dtypes: if not None, it's used to enforce the dtype of the new levels.
                It can be a single dtype, or a list of dtypes, one for each condition.
                Examples: int, float, "category"...

        Returns:
            resulting Series or DataFrame.
        """
        conditions = ensure_list(conditions)
        values = ensure_list(values)
        if len(conditions) != len(values):
            raise ValueError("Conditions and values must have the same length")
        result = smart_concat([self._obj], keys=[tuple(values)], names=conditions)
        if dtypes:
            dtypes = ensure_list(dtypes)
            if len(conditions) != len(dtypes):
                raise ValueError("Conditions and dtypes must have the same length")
            result.index = result.index.etl.astype(dict(zip(conditions, dtypes)))
        if drop:
            # levels to be dropped, for example: [-3, -2, -1]
            levels = list(range(-self._obj.index.nlevels, 0))
            result = result.droplevel(levels)
        elif inner:
            # rotate the levels: (0 1) 2 3 4 5 -> 2 3 4 5 (0 1)
            order = list(range(result.index.nlevels))
            order = order[len(conditions) :] + order[: len(conditions)]
            result = result.reorder_levels(order)
        return result

    def select(self, drop_level: bool = True, **kwargs) -> PandasT:
        """Filter the series or dataframe based on some conditions on the index.

        Note: if the level doesn't need to be dropped, it's possible to use `etl.q` instead.
        TODO: consider if it can be deprecated in favour of etl.q, and removed.

        Args:
            drop_level (bool): True to drop the conditions from the returned object.
            kwargs: conditions used to filter, specified as name=value.
        """
        if not kwargs:
            return self._obj
        labels, values = zip(*kwargs.items())
        return self._obj.xs(level=labels, key=values, drop_level=drop_level, axis=0)

    def groupby_except(self, conditions: Union[str, list[str]], *args, **kwargs) -> PandasGroupByT:
        """Group by all the conditions except for the ones specified.

        Args:
            conditions: single condition or list of conditions to be excluded from the groupby
            args: positional arguments to be passed to groupby
            kwargs: key arguments to be passed to groupby
        """
        complementary_conditions = self.complementary_conditions(conditions)
        return self._obj.groupby(complementary_conditions, *args, **kwargs)

    def pool(self, conditions: Union[str, list[str]], func: Callable) -> PandasT:
        """Remove one or more conditions grouping by the remaining conditions.

        Args:
            conditions: single condition or list of conditions to be removed from the index.
            func: function that should accept a single element.
                If the returned value is a Series, it will be used as an additional level
                in the MultiIndex of the returned object.
        """
        return self.groupby_except(conditions).apply(func)

    @abstractmethod
    def _query_list(self, query_list: list[dict[str, Any]]) -> PandasT:
        """Given a list of query dictionaries, return the filtered Series or DataFrame."""

    def q(
        self, _query: Optional[Union[dict[str, Any], list[dict[str, Any]]]] = None, /, **params
    ) -> PandasT:
        """Given a query dict, list, or some query params, return the filtered Series or DataFrame.

        Filter by columns (for DataFrames) and index names (for both Series and DataFrames).
        If a name is present in both columns and index names, only the column is considered.

        If a list is passed, the items must be query dictionaries, that will be OR-ed together.

        All the keys of the dict are combined in AND, while the values can be scalar, list, or dict.

        * If value is a scalar, the exact value will be matched.
        * If value is a list, the values in the list are considered in OR.
        * If value is a dict, a more advanced filter can be specified using the
            supported operators: ``eq, ne, le, lt, ge, gt, isin``.

        Query and named params cannot be specified together.
        If they are both empty or missing, the original Series or DataFrame is returned.

        This method is similar to the standard `query` method for pandas DataFrames, but it accepts
        a dict instead of a string, and has some limitations on the possible filters to be applied.

        Args:
            _query: query dictionary, where the keys are columns or index levels, and the values
                can be scalar, list, or dict values.
            **params: named params can be specified as an alternative to the _query dictionary.

        Examples:
            * {"mtype": "SO_BP", "etype": "cNAC"} -> mtype == SO_BP AND etype == cNAC
            * {"mtype": ["SO_BP", "SP_AA"]} -> mtype == SO_BP OR mtype == SP_AA
            * {"gid": {"ge": 3, "lt": 8} -> gid >= 3 AND gid < 8
        """
        if _query and params:
            raise ValueError("Query and params cannot be specified together")
        return self._query_list(ensure_list(_query or params))

    def one(self, _query: Optional[dict] = None, /, **params) -> Any:
        """Execute the query and return the unique resulting record."""
        result = self.q(_query, **params)
        if len(result) != 1:
            raise RuntimeError(f"The query returned {len(result)} records instead of 1.")
        return result.iloc[0]

    def first(self, _query: Optional[dict] = None, /, **params) -> Any:
        """Execute the query and return the first resulting record."""
        result = self.q(_query, **params)
        if len(result) == 0:
            raise RuntimeError("The query returned 0 records.")
        return result.iloc[0]


class ETLSeriesAccessor(ETLBaseAccessor[pd.Series, SeriesGroupBy]):
    """Series accessor."""

    def unpool(self, func: Callable) -> pd.Series:
        """Apply the given function to the object elements and add a condition to the index.

        Args:
            func: function that should accept a single element and return a Series object.
                The name of that Series will be used as the name of the new level
                in the MultiIndex of the returned object.
        """
        return self._obj.apply(func).stack()

    def iter(self) -> Iterator[tuple[NamedTuple, Any]]:
        """Iterate over the items, yielding a tuple (named_index, value) for each element.

        The returned named_index is a namedtuple representing the value of the index.
        The returned value is the actual value of each element of the series.
        """
        return zip(self._obj.index.etl.iter(), iter(self._obj))

    def iterdict(self) -> Iterator[tuple[dict, Any]]:
        """Iterate over the items, yielding a tuple (named_index, value) for each element.

        The returned named_index is a dict representing the value of the index.
        The returned value is the actual value of each element of the series.

        This method can be used as an alternative to ``iter`` when:

        - The index names contain invalid identifiers, or
        - it's more convenient to work with dictionaries.

        Valid identifiers consist of letters, digits, and underscores but do not start
        with a digit or underscore and cannot be a Python keyword.
        """
        return zip(self._obj.index.etl.iterdict(), iter(self._obj))

    def _query_list(self, query_list: list[dict[str, Any]]) -> pd.Series:
        """Given a list of query dictionaries, return the Series filtered by index."""
        return query_series(self._obj, query_list)


class ETLDataFrameAccessor(ETLBaseAccessor[pd.DataFrame, DataFrameGroupBy]):
    """DataFrame accessor."""

    def iter(self) -> Iterator[tuple[NamedTuple, NamedTuple]]:
        """Iterate over the items, yielding a tuple (named_index, value) for each element.

        The returned ``named_index`` is a namedtuple representing the value of the index.
        The returned ``value`` is a namedtuple as returned by pandas.DataFrame.itertuples.
        """
        return zip(self._obj.index.etl.iter(), self._obj.itertuples(index=False, name="Values"))

    def iterdict(self) -> Iterator[tuple[dict, dict]]:
        """Iterate over the items, yielding a tuple (named_index, value) for each element.

        The returned ``named_index`` is a dict representing the value of the index.
        The returned ``value`` is a dict containing a key for each column.

        This method can be used as an alternative to ``iter`` when:

        - The column or index names contain invalid identifiers, or
        - it's more convenient to work with dictionaries.

        Valid identifiers consist of letters, digits, and underscores but do not start
        with a digit or underscore and cannot be a Python keyword.
        """
        columns = self._obj.columns
        for named_index, value in zip(
            self._obj.index.etl.iterdict(), self._obj.itertuples(index=False)
        ):
            yield named_index, dict(zip(columns, value))

    def _query_list(self, query_list: list[dict[str, Any]]) -> pd.DataFrame:
        """Given a list of query dicts, return the DataFrame filtered by columns and index."""
        return query_frame(self._obj, query_list)

    def groupby_iter(
        self,
        groupby_columns: list[str],
        selected_columns: Optional[list[str]] = None,
        sort: bool = True,
        observed: bool = True,
    ) -> Iterator[tuple[tuple, pd.DataFrame]]:
        """Group the dataframe by columns and yield each record as a tuple (key, df).

        It can be used as a replacement for the iteration over df.groupby, but:
            - the yielded keys are namedtuples, instead of tuples
            - the yielded dataframes contain only the selected columns, if specified

        Args:
            groupby_columns: list of column names to group by.
            selected_columns: list of column names to be included in the yielded dataframes.
                If None, all the columns are included.
            sort: Sort group keys.
            observed: This only applies if any of the groupers are Categoricals.
                If True: only show observed values for categorical groupers.
                If False: show all values for categorical groupers.

        Yields:
            a tuple (key, df), where key is a namedtuple with the grouped columns
        """
        # Workaround to avoid: FutureWarning: In a future version of pandas, a length 1 tuple will
        # be returned when iterating over a groupby with a grouper equal to a list of length 1.
        # Don't supply a list with a single grouper to avoid this warning.
        # See also https://github.com/pandas-dev/pandas/issues/53500
        fix_pandas = len(groupby_columns) == 1 and (
            Version(pd.__version__) < Version("2")
            or Version(pd.__version__) < Version("2.1")
            and selected_columns
        )
        by = groupby_columns[0] if fix_pandas else groupby_columns
        grouped = self._obj.groupby(by, sort=sort, observed=observed)
        if selected_columns:
            grouped = grouped[selected_columns]
        RecordKey = namedtuple("RecordKey", groupby_columns)  # type: ignore[misc]
        if fix_pandas:
            for key, df in grouped:
                yield RecordKey(key), df  # type: ignore[call-arg]
        else:
            for key, df in grouped:
                yield RecordKey(*key), df

    def groupby_run_parallel(
        self,
        groupby_columns: list[str],
        selected_columns: Optional[list[str]] = None,
        *,
        sort: bool = True,
        observed: bool = True,
        func: Optional[Callable] = None,
        jobs: Optional[int] = None,
        backend: Optional[str] = None,
    ) -> list[Any]:
        """Call groupby_iter and apply the given function in parallel, returning the results.

        Args:
            groupby_columns: see groupby_iter.
            selected_columns: see groupby_iter.
            sort: see groupby_iter.
            observed: see groupby_iter.
            func: callable accepting the parameters: key (NamedTuple), df (pd.DataFrame)
            jobs: number of jobs (see run_parallel)
            backend: parallel backend (see run_parallel)

        Returns:
            list of results.
        """
        assert func is not None, "A callable must be specified."
        grouped = self.groupby_iter(
            groupby_columns=groupby_columns,
            selected_columns=selected_columns,
            sort=sort,
            observed=observed,
        )
        tasks_generator = (Task(partial(func, key=key, df=group_df)) for key, group_df in grouped)
        return run_parallel(tasks_generator, jobs=jobs, backend=backend)

    def groupby_apply_parallel(
        self,
        groupby_columns: list[str],
        selected_columns: Optional[list[str]] = None,
        *,
        sort: bool = True,
        observed: bool = True,
        func: Optional[Callable] = None,
        jobs: Optional[int] = None,
        backend: Optional[str] = None,
    ) -> pd.DataFrame:
        """Call groupby_iter and apply the given function in parallel, returning a DataFrame.

        To work as expected, func should return a DataFrame or a Series, and all the returned
        objects should have the same index and columns.

        Still experimental.
        """
        results = self.groupby_run_parallel(
            groupby_columns=groupby_columns,
            selected_columns=selected_columns,
            sort=sort,
            observed=observed,
            func=func,
            jobs=jobs,
            backend=backend,
        )
        return smart_concat(results)


class ETLIndexAccessor:
    """Index accessor."""

    def __init__(self, pandas_obj: pd.Index) -> None:
        """Initialize the accessor."""
        self._obj = pandas_obj

    def _mangle_names(self) -> list[str]:
        """Return the index names, replacing missing names with ilevel_N, where N is the level.

        The name ilevel_N is the same name used by convention in DataFrame.query to identify
        the index levels without a name in a MultiIndex.
        """
        return [
            name if name is not None else f"ilevel_{i}" for i, name in enumerate(self._obj.names)
        ]

    def iter(self) -> Iterator[tuple]:
        """Iterate over the index, yielding a namedtuple for each element.

        It can be used as an alternative to the pandas iteration over the index
        to yield named tuples instead of standard tuples.

        It works with both Indexes and MultiIndexes.
        """
        names = self._mangle_names()
        Index = namedtuple("Index", names, rename=True)  # type: ignore[misc]
        if len(names) == 1:
            for i in self._obj:
                yield Index(i)  # type: ignore[call-arg]
        else:
            for i in self._obj:
                yield Index(*i)

    def iterdict(self) -> Iterator[dict]:
        """Iterate over the index, yielding a dict for each element.

        This method can be used as an alternative to ``iter`` when:

        - The index names contain invalid identifiers, or
        - it's more convenient to work with dictionaries.

        It works with both Indexes and MultiIndexes.
        """
        names = self._mangle_names()
        if len(names) == 1:
            for i in self._obj:
                yield {names[0]: i}
        else:
            for i in self._obj:
                yield dict(zip(names, i))

    @property
    def dtypes(self) -> pd.Series:
        """Return the dtypes of the index."""
        if isinstance(self._obj, pd.MultiIndex):
            return self._obj.dtypes
        return pd.Series([self._obj.dtype], index=pd.Index([self._obj.name]))

    def astype(self, dtype) -> pd.Index:
        """Create a new Index with the given dtypes.

        Args:
            dtype: numpy dtype or pandas type, or dict of dtypes when applied to a MultiIndex.
                Any (u)int16 or (u)int32 dtype is considered as (u)int64,
                since Pandas doesn't have a corresponding Index type for them.

        Returns:
            a copy of index using the specified dtypes.
        """
        if not isinstance(self._obj, pd.MultiIndex):
            return self._obj.astype(dtype)
        if not isinstance(dtype, dict):
            raise TypeError("A dict of dtypes must be specified when working with a MultiIndex")
        if diff := set(dtype).difference(self._obj.names):
            raise NameError(f"Some names don't exist in the index levels: {sorted(diff)}")
        dtype_by_num = {i: dtype[name] for i, name in enumerate(self._obj.names) if name in dtype}
        return self._obj.set_levels(
            [self._obj.levels[i].astype(dt) for i, dt in dtype_by_num.items()],
            level=list(dtype_by_num),
        )


def register_accessors() -> None:
    """Register the accessors.

    It must be called once, before accessing the etl namespace.
    """
    L.info("Registering etl pandas accessors")
    # alternatively, the accessors could be registered with a decorator,
    # but this call is more explicit and it doesn't need any unused import
    pd.api.extensions.register_series_accessor("etl")(ETLSeriesAccessor)
    pd.api.extensions.register_dataframe_accessor("etl")(ETLDataFrameAccessor)
    pd.api.extensions.register_index_accessor("etl")(ETLIndexAccessor)



from typing import Iterator

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal, assert_equal
from pandas.testing import assert_frame_equal, assert_index_equal, assert_series_equal

from blueetl_core.etl import ETLDataFrameAccessor


def test_etl_instance(dataframe1):
    obj = dataframe1
    result = obj.etl
    assert isinstance(result, ETLDataFrameAccessor)


def test_conditions(dataframe1):
    obj = dataframe1
    result = obj.etl.conditions()
    assert_array_equal(result, ["i0", "i1"])


@pytest.mark.parametrize(
    "conditions, expected",
    [
        pytest.param("i0", ["i1"], id="single"),
        pytest.param(["i0"], ["i1"], id="list"),
        pytest.param([], ["i0", "i1"], id="empty"),
        pytest.param(["i0", "i1"], [], id="all"),
        pytest.param(["iX"], ["i0", "i1"], id="nonexistent"),
    ],
)
def test_complementary_conditions(dataframe1, conditions, expected):
    obj = dataframe1
    result = obj.etl.complementary_conditions(conditions)
    assert_array_equal(result, expected)


@pytest.mark.parametrize(
    "conditions, expected",
    [
        pytest.param(None, [pd.Index(["a", "b"], name="i0"), pd.Index(["c", "d"], name="i1")]),
        pytest.param(["i0"], [pd.Index(["a", "b"], name="i0")], id="one"),
    ],
)
def test_labels(dataframe1, conditions, expected):
    obj = dataframe1
    result = obj.etl.labels(conditions)
    assert len(result) == len(expected)
    for actual_index, expected_index in zip(result, expected):
        assert_index_equal(actual_index, expected_index)


def test_labels_of(dataframe1):
    obj = dataframe1
    result = obj.etl.labels_of("i0")
    assert_array_equal(result, ["a", "b"])


def test_labels_of_nonexistent_condition(dataframe1):
    obj = dataframe1
    with pytest.raises(KeyError, match="Level iX not found"):
        obj.etl.labels_of("iX")


@pytest.mark.parametrize(
    "conditions",
    [
        pytest.param("i0", id="str"),
        pytest.param(["i0"], id="list"),
    ],
)
def test_remove_conditions(dataframe1, conditions):
    obj = dataframe1
    result = obj.etl.remove_conditions(conditions)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0": [0, 1, 2, 3], "v1": [4, 5, 6, 7]},
            index=pd.Index(["c", "d", "c", "d"], name="i1"),
        ),
    )


@pytest.mark.parametrize(
    "conditions",
    [
        pytest.param("i1", id="str"),
        pytest.param(["i1"], id="list"),
    ],
)
def test_keep_conditions(dataframe1, conditions):
    obj = dataframe1
    result = obj.etl.keep_conditions(conditions)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0": [0, 1, 2, 3], "v1": [4, 5, 6, 7]},
            index=pd.Index(["c", "d", "c", "d"], name="i1"),
        ),
    )


@pytest.mark.parametrize(
    "conditions, values",
    [
        pytest.param("i2", "e", id="str"),
        pytest.param(["i2"], ["e"], id="list"),
    ],
)
def test_add_conditions(dataframe1, conditions, values):
    obj = dataframe1
    result = obj.etl.add_conditions(conditions, values)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0": [0, 1, 2, 3], "v1": [4, 5, 6, 7]},
            index=pd.MultiIndex.from_tuples(
                [("e", "a", "c"), ("e", "a", "d"), ("e", "b", "c"), ("e", "b", "d")],
                names=["i2", "i0", "i1"],
            ),
        ),
    )


@pytest.mark.parametrize(
    "conditions, values, dtypes",
    [
        pytest.param("i2", 123, float, id="str"),
        pytest.param(["i2"], [123], [float], id="list"),
    ],
)
def test_add_conditions_with_dtypes(dataframe1, conditions, values, dtypes):
    obj = dataframe1
    result = obj.etl.add_conditions(conditions, values, dtypes=dtypes)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0": [0, 1, 2, 3], "v1": [4, 5, 6, 7]},
            index=pd.MultiIndex.from_tuples(
                [(123.0, "a", "c"), (123.0, "a", "d"), (123.0, "b", "c"), (123.0, "b", "d")],
                names=["i2", "i0", "i1"],
            ),
        ),
    )
    assert_equal(result.index.dtypes.to_numpy(), [float, object, object])


@pytest.mark.parametrize(
    "inner, drop, expected_conditions",
    [
        (False, False, ["i2", "i0", "i1"]),
        (True, False, ["i0", "i1", "i2"]),
        (False, True, ["i2"]),
        (True, True, ["i2"]),
    ],
)
def test_add_conditions_inner_drop(dataframe1, inner, drop, expected_conditions):
    obj = dataframe1
    conditions = ["i2"]
    values = ["e"]
    result = obj.etl.add_conditions(conditions, values, inner=inner, drop=drop)
    assert result.index.names == expected_conditions


def test_select_with_drop_level(dataframe1):
    obj = dataframe1
    result = obj.etl.select(i1="c", drop_level=True)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0": [0, 2], "v1": [4, 6]},
            index=pd.Index(["a", "b"], name="i0"),
        ),
    )


def test_select_without_drop_level(dataframe1):
    obj = dataframe1
    result = obj.etl.select(i1="c", drop_level=False)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0": [0, 2], "v1": [4, 6]},
            index=pd.MultiIndex.from_tuples([("a", "c"), ("b", "c")], names=["i0", "i1"]),
        ),
    )


def test_select_without_args(dataframe1):
    obj = dataframe1
    result = obj.etl.select()
    assert_frame_equal(result, obj)


@pytest.mark.skip(reason="need to verify the requirements and the implementation")
def test_unpool_from_series_of_scalars(dataframe1):
    def func(x):
        return pd.Series([x, x + 20], index=pd.Index(["x0", "x2"], name="func"))

    obj = dataframe1
    result = obj.etl.unpool(func)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0": [0, 20, 1, 21, 2, 22, 3, 23], "v1": [4, 24, 5, 25, 6, 26, 7, 27]},
            index=pd.MultiIndex.from_tuples(
                [
                    ("a", "c", "x0"),
                    ("a", "c", "x2"),
                    ("a", "d", "x0"),
                    ("a", "d", "x2"),
                    ("b", "c", "x0"),
                    ("b", "c", "x2"),
                    ("b", "d", "x0"),
                    ("b", "d", "x2"),
                ],
                names=["i0", "i1", "func"],
            ),
        ),
    )


@pytest.mark.skip(reason="need to verify the requirements and the implementation")
def test_unpool_from_series_of_lists(dataframe1):
    def func(x):
        return pd.Series([x[0], x[2]], index=pd.Index(["x0", "x2"], name="func"))

    obj = dataframe1
    # transform each element to a list of integers
    obj = obj.transform(lambda x: [x, x + 10, x + 20])
    result = obj.etl.unpool(func)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0": [0, 1, 2, 3], "v1": [4, 5, 6, 7]},
            # [0, 20, 1, 21, 2, 22, 3, 23],
            index=pd.MultiIndex.from_tuples(
                [
                    ("a", "c", "x0"),
                    ("a", "c", "x2"),
                    ("a", "d", "x0"),
                    ("a", "d", "x2"),
                    ("b", "c", "x0"),
                    ("b", "c", "x2"),
                    ("b", "d", "x0"),
                    ("b", "d", "x2"),
                ],
                names=["i0", "i1", "func"],
            ),
        ),
    )


def test_pool_with_func_returning_scalar(dataframe1):
    def func(x):
        # x is a dataframe with the same index as the original obj,
        # but the only varying level is the one specified in the pool call.
        # Example:
        #        v0  v1
        # i0 i1
        # a  c    0   4
        #    d    1   5
        return x.to_numpy().sum()

    obj = dataframe1
    result = obj.etl.pool(["i1"], func)
    assert_series_equal(
        result,
        pd.Series(
            [10, 18],
            index=pd.Index(["a", "b"], name="i0"),
        ),
    )


def test_pool_with_func_returning_series(dataframe1):
    def func(x):
        # x is a dataframe with the same index as the original obj,
        # but the only varying level is the one specified in the pool call.
        # Example:
        #        v0  v1
        # i0 i1
        # a  c    0   4
        #    d    1   5
        return x.sum()

    obj = dataframe1
    result = obj.etl.pool(["i1"], func)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0": [1, 5], "v1": [9, 13]},
            index=pd.Index(["a", "b"], name="i0"),
        ),
    )


def test_pool_with_func_returning_series_with_different_shape(dataframe1):
    def func(x):
        # x is a dataframe with the same index as the original obj,
        # but the only varying level is the one specified in the pool call.
        # Example:
        #        v0  v1
        # i0 i1
        # a  c    0   4
        #    d    1   5
        return pd.Series(
            [x["v0"].sum(), x["v1"].sum(), x.to_numpy().sum()],
            index=pd.Index(["v0_sum", "v1_sum", "sum"]),
        )

    obj = dataframe1
    result = obj.etl.pool(["i1"], func)
    assert_frame_equal(
        result,
        pd.DataFrame(
            {"v0_sum": [1, 5], "v1_sum": [9, 13], "sum": [10, 18]},
            index=pd.Index(["a", "b"], name="i0"),
        ),
    )


@pytest.mark.parametrize("as_type", ["dict", "list", "kwargs"])
@pytest.mark.parametrize(
    "query, expected",
    [
        (
            {"v0": 0},
            pd.DataFrame(
                [[0, 4]],
                columns=["v0", "v1"],
                index=pd.MultiIndex.from_tuples([("a", "c")], names=["i0", "i1"]),
            ),
        ),
        (
            {"v0": 1, "v1": 5},
            pd.DataFrame(
                [[1, 5]],
                columns=["v0", "v1"],
                index=pd.MultiIndex.from_tuples([("a", "d")], names=["i0", "i1"]),
            ),
        ),
        (
            {"v0": 1, "v1": 8},
            pd.DataFrame(
                columns=["v0", "v1"],
                index=pd.MultiIndex.from_tuples([], names=["i0", "i1"]),
                dtype=np.int64,
            ),
        ),
        (
            {"v0": [1, 2]},
            pd.DataFrame(
                [[1, 5], [2, 6]],
                columns=["v0", "v1"],
                index=pd.MultiIndex.from_tuples([("a", "d"), ("b", "c")], names=["i0", "i1"]),
            ),
        ),
        (
            {"v0": [1, 2], "v1": [6, 7]},
            pd.DataFrame(
                [[2, 6]],
                columns=["v0", "v1"],
                index=pd.MultiIndex.from_tuples([("b", "c")], names=["i0", "i1"]),
            ),
        ),
        (
            {"v0": {"ge": 1, "le": 2}, "v1": {"isin": [6, 7]}},
            pd.DataFrame(
                [[2, 6]],
                columns=["v0", "v1"],
                index=pd.MultiIndex.from_tuples([("b", "c")], names=["i0", "i1"]),
            ),
        ),
    ],
)
def test_query(dataframe1, query, expected, as_type):
    obj = dataframe1
    assert isinstance(query, dict)

    if as_type == "dict":
        result = obj.etl.q(query)
    elif as_type == "list":
        result = obj.etl.q([query])
    elif as_type == "kwargs":
        result = obj.etl.q(**query)
    else:
        raise ValueError(f"Invalid as_type: {as_type}")

    # check_index_type=False needed to avoid Attribute "inferred_type" are different
    assert_frame_equal(result, expected, check_index_type=len(expected) != 0)


def test_query_with_multiple_dicts(dataframe1):
    obj = dataframe1
    query = [
        {"v0": 0},
        {"v0": 1, "v1": 5},
    ]
    expected = pd.DataFrame(
        [[0, 4], [1, 5]],
        columns=["v0", "v1"],
        index=pd.MultiIndex.from_tuples([("a", "c"), ("a", "d")], names=["i0", "i1"]),
    )

    result = obj.etl.q(query)

    assert_frame_equal(result, expected)


def test_query_with_invalid_params(dataframe1):
    obj = dataframe1
    with pytest.raises(ValueError, match="Query and params cannot be specified together"):
        obj.etl.q({"v0": 0}, v1=1)


def test_one(dataframe1):
    obj = dataframe1
    result = obj.etl.one(v0=0)
    assert_series_equal(result, pd.Series({"v0": 0, "v1": 4}, name=("a", "c")))


def test_one_raises_when_multiple_results(dataframe1):
    obj = dataframe1
    with pytest.raises(RuntimeError, match="The query returned 2 records instead of 1."):
        obj.etl.one(i0="b")


def test_one_raises_when_no_results(dataframe1):
    obj = dataframe1
    with pytest.raises(RuntimeError, match="The query returned 0 records instead of 1."):
        obj.etl.one(i0="x")


def test_first(dataframe1):
    obj = dataframe1
    result = obj.etl.first(i0="b")
    assert_series_equal(result, pd.Series({"v0": 2, "v1": 6}, name=("b", "c")))


def test_first_raises_when_no_results(dataframe1):
    obj = dataframe1
    with pytest.raises(RuntimeError, match="The query returned 0 records."):
        obj.etl.first(i0="x")


def test_iter(dataframe1):
    obj = dataframe1
    it = obj.etl.iter()
    assert isinstance(it, Iterator)
    index, value = next(it)
    assert isinstance(index, tuple)
    assert isinstance(value, tuple)
    assert index == ("a", "c")
    assert value == (0, 4)
    assert index.i0 == "a"
    assert index.i1 == "c"
    assert value.v0 == 0
    assert value.v1 == 4


def test_iterdict(dataframe1):
    obj = dataframe1
    it = obj.etl.iterdict()
    assert isinstance(it, Iterator)
    index, value = next(it)
    assert isinstance(index, dict)
    assert isinstance(value, dict)
    assert index == {"i0": "a", "i1": "c"}
    assert value == {"v0": 0, "v1": 4}


@pytest.mark.parametrize(
    "params, expected_key, expected_df",
    [
        pytest.param(
            {"groupby_columns": ["i1"]},
            ("c",),
            pd.DataFrame(
                [[0, 4], [2, 6]],
                columns=["v0", "v1"],
                index=pd.MultiIndex.from_tuples([("a", "c"), ("b", "c")], names=["i0", "i1"]),
            ),
        ),
        pytest.param(
            {"groupby_columns": ["i1"], "selected_columns": ["v1"]},
            ("c",),
            pd.DataFrame(
                [[4], [6]],
                columns=["v1"],
                index=pd.MultiIndex.from_tuples([("a", "c"), ("b", "c")], names=["i0", "i1"]),
            ),
        ),
        pytest.param(
            {"groupby_columns": ["v0"]},
            (0,),
            pd.DataFrame(
                [[0, 4]],
                columns=["v0", "v1"],
                index=pd.MultiIndex.from_tuples([("a", "c")], names=["i0", "i1"]),
            ),
        ),
        pytest.param(
            {"groupby_columns": ["v0"], "selected_columns": ["v1"]},
            (0,),
            pd.DataFrame(
                [[4]],
                columns=["v1"],
                index=pd.MultiIndex.from_tuples([("a", "c")], names=["i0", "i1"]),
            ),
        ),
    ],
)
def test_groupby_iter(dataframe1, params, expected_key, expected_df):
    obj = dataframe1
    it = obj.etl.groupby_iter(**params)
    assert isinstance(it, Iterator)
    key, group_df = next(it)
    assert key == expected_key
    assert_frame_equal(group_df, expected_df)


@pytest.mark.parametrize(
    "params, expected",
    [
        (
            {"groupby_columns": ["i0"]},
            [
                (("a",), [{"v0": 0, "v1": 4}, {"v0": 1, "v1": 5}]),
                (("b",), [{"v0": 2, "v1": 6}, {"v0": 3, "v1": 7}]),
            ],
        ),
        (
            {"groupby_columns": ["v0"]},
            [
                ((0,), [{"v0": 0, "v1": 4}]),
                ((1,), [{"v0": 1, "v1": 5}]),
                ((2,), [{"v0": 2, "v1": 6}]),
                ((3,), [{"v0": 3, "v1": 7}]),
            ],
        ),
        (
            {"groupby_columns": ["i0", "v0"]},
            [
                (("a", 0), [{"v0": 0, "v1": 4}]),
                (("a", 1), [{"v0": 1, "v1": 5}]),
                (("b", 2), [{"v0": 2, "v1": 6}]),
                (("b", 3), [{"v0": 3, "v1": 7}]),
            ],
        ),
        (
            {"groupby_columns": ["i0", "v0"], "selected_columns": ["v1"]},
            [
                (("a", 0), [{"v1": 4}]),
                (("a", 1), [{"v1": 5}]),
                (("b", 2), [{"v1": 6}]),
                (("b", 3), [{"v1": 7}]),
            ],
        ),
    ],
)
def test_groupby_run_parallel(dataframe1, params, expected):
    def func(key, df):
        return tuple(key), df.to_dict(orient="records")

    obj = dataframe1
    result = obj.etl.groupby_run_parallel(**params, func=func)
    assert result == expected


@pytest.mark.parametrize(
    "params, expected",
    [
        (
            {"groupby_columns": ["i0"]},
            pd.Series(
                [4, 6, 8, 10],
                index=pd.MultiIndex.from_tuples(
                    [("a", "c"), ("a", "d"), ("b", "c"), ("b", "d")], names=["i0", "i1"]
                ),
            ),
        ),
    ],
)
def test_groupby_apply_parallel(dataframe1, params, expected):
    def func(key, df):
        return df.sum(axis=1)

    obj = dataframe1
    result = obj.etl.groupby_apply_parallel(**params, func=func)
    assert_series_equal(result, expected)



"""BlueETL core transformations."""
import logging
from typing import Optional, Union

L = logging.getLogger(__name__)


def setup_logging(loglevel: Union[int, str], logformat: Optional[str] = None, **logparams) -> None:
    """Setup logging."""
    logformat = logformat or "%(asctime)s %(levelname)s %(name)s: %(message)s"
    logging.basicConfig(format=logformat, level=loglevel, **logparams)



"""Utilities for parallelization."""
import logging
import os
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any, Callable, Optional

import numpy as np
from joblib import Parallel, delayed

from blueetl_core.constants import (
    BLUEETL_JOBLIB_BACKEND,
    BLUEETL_JOBLIB_JOBS,
    BLUEETL_JOBLIB_VERBOSE,
    BLUEETL_SUBPROCESS_LOGGING_LEVEL,
)
from blueetl_core.logging import L, setup_logging


@dataclass
class TaskContext:
    """TaskContext class containing information to be passed to the tasks in subprocesses."""

    task_id: int
    loglevel: int
    seed: Optional[int] = None
    ppid: Optional[int] = None


class Task:
    """Task class."""

    def __init__(self, func: Callable) -> None:
        """Initialize the Task object.

        Args:
            func: function to be executed whe the Task is called in a subprocess.
        """
        self.func = func

    def _setup_logging(self, ctx: TaskContext) -> None:
        """Initialize logging in a subprocess."""
        loglevel = os.getenv(BLUEETL_SUBPROCESS_LOGGING_LEVEL)
        if loglevel:
            logformat = f"%(asctime)s %(levelname)s %(name)s [task={ctx.task_id}]: %(message)s"
            setup_logging(loglevel=loglevel, logformat=logformat, force=True)

    @staticmethod
    def _setup_seed(ctx: TaskContext) -> None:
        """Initialize seed in a subprocess."""
        if ctx.seed is not None:
            np.random.seed(ctx.seed)

    def __call__(self, ctx: TaskContext) -> Any:
        """Call the wrapped function and return the result.

        It's intended to be called in a subprocess.

        Args:
            ctx: TaskContext instance containing the context information.
        """
        if ctx.ppid and ctx.ppid != os.getpid():
            # only when the task is executed in a subprocess
            self._setup_logging(ctx)
            self._setup_seed(ctx)
        return self.func()


def run_parallel(
    tasks: Iterable[Callable[[TaskContext], Any]],
    jobs: Optional[int] = None,
    backend: Optional[str] = None,
    verbose: Optional[int] = None,
    base_seed: Optional[int] = None,
) -> list[Any]:
    """Run tasks in parallel.

    Args:
        tasks: iterable of callable objects that will be called in separate threads or processes.
            The callable must accept a single parameter ctx, that will contain a TaskContext.
        jobs: number of jobs. If not specified, use the BLUEETL_JOBLIB_JOBS env variable,
            or use half of the available cpus. Set to 1 to disable parallelization.
        backend: backend passed to joblib. If not specified, use the BLUEETL_JOBLIB_BACKEND env
            variable, or use the joblib default (loky).
            Possible values: loky, multiprocessing, threading.
        verbose: verbosity of joblib. If not specified, use the BLUEETL_JOBLIB_VERBOSE.
        base_seed: initial base seed. If specified, a different seed is added to the task context,
            and passed to each callable object.

    Returns:
        list of objects returned by the callable objects, in the same order.
    """
    loglevel = L.getEffectiveLevel()
    if verbose is None:
        verbose_env = os.getenv(BLUEETL_JOBLIB_VERBOSE)
        verbose = int(verbose_env) if verbose_env else 0 if loglevel >= logging.WARNING else 10
    if not jobs:
        jobs_env = os.getenv(BLUEETL_JOBLIB_JOBS)
        jobs = int(jobs_env) if jobs_env else max((os.cpu_count() or 1) // 2, 1)
    if not backend:
        backend = os.getenv(BLUEETL_JOBLIB_BACKEND)
    parallel = Parallel(n_jobs=jobs, backend=backend, verbose=verbose)
    return parallel(
        delayed(task)(
            ctx=TaskContext(
                task_id=i,
                loglevel=loglevel,
                seed=None if base_seed is None else base_seed + i,
                ppid=os.getpid(),
            )
        )
        for i, task in enumerate(tasks)
    )
